(**********************************************************************)
(* Equations                                                          *)
(* Copyright (c) 2009-2016 Matthieu Sozeau <matthieu.sozeau@inria.fr> *)
(**********************************************************************)
(* This file is distributed under the terms of the                    *)
(* GNU Lesser General Public License Version 2.1                      *)
(**********************************************************************)

Require Import Program Bvector List Relations.
From Equations Require Import Equations Signature.
From Equations Require Import DepElimDec EqDec.

Require Import Utf8.

Ltac rec ::= Subterm.rec_wf_eqns.

Ltac solve_rec ::= simpl in * ; cbv zeta ; intros ;
  try typeclasses eauto with subterm_relation Below lt.

Require Import Vectors.Vector.

Arguments Vector.nil {A}.
Arguments Vector.cons {A} _ {n}.
Notation " x :: y " :=  (@Vector.cons _ x _ y) (at level 60, right associativity) : vect_scope.
Notation " x |: n :| y " := (@Vector.cons _ x n y) (at level 20, right associativity) : vect_scope.
Notation " [ ] " := Vector.nil (at level 0) : vect_scope.
Local Open Scope vect_scope.

(** Concatenation *)

Equations vapp {A} {n m} (v : vector A n) (w : vector A m) : vector A (n + m) :=
  vapp [] w := w ;
  vapp (Vector.cons a n v) w := a :: vapp v w.
Infix "++" := vapp : vect_scope.

Derive Signature Subterm for vector.
Print Assumptions well_founded_t_subterm.
Notation vector_subterm A := (t_subterm A).

Section Unzip.
  Context {A B : Type}.

  Equations unzip {n} (v : vector (A * B) n) : vector A n * vector B n :=
  unzip v by rec (signature_pack v) (@vector_subterm (A * B)) :=
  unzip nil := ([], []) ;
  unzip (cons (pair x y) n v) with unzip v := {
    | pair xs ys := (x :: xs, y :: ys) }.

End Unzip.

Print Assumptions unzip.
Extraction unzip.

Inductive Split {X : Type}{m n : nat} : vector X (m + n) -> Type :=
  append : ∀ (xs : vector X m) (ys : vector X n), Split (vapp xs ys).

Arguments Split [ X ].

(** In some cases, the index can be relevant though: *)

Equations split {X : Type} {m n} (xs : vector X (m + n)) : Split m n xs :=
split {m:=m} xs by rec m :=
split {m:=0} xs := append nil xs ;
split {m:=(S m')} {n:=n} (cons x ?(m' + n) xs) <= split xs => {
  | append xs' ys' := append (x :: xs') ys' }.

Print Assumptions split.
Extraction split.

Lemma split_vapp' : ∀ (X : Type) m n (v : vector X m) (w : vector X n),
  let 'append v' w' := split (v ++ w) in
    v = v' /\ w = w'.
Proof.
  intros.
  funelim (vapp v w).
  destruct split. depelim xs; intuition.
  simp split in *. destruct split. simpl.
  intuition congruence.
Qed.
